import 'dart:io';

import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:firebase_storage/firebase_storage.dart';

class FirebaseRepository {
  final Firestore _firestoreClient;
  final FirebaseStorage _firestorageClient;

  FirebaseRepository({
    Firestore firestoreClient,
    FirebaseStorage firestorageClient,
  })  : _firestoreClient = firestoreClient ?? Firestore.instance,
        _firestorageClient = firestorageClient ?? FirebaseStorage.instance;

  Future<dynamic> uploadMedia(File file, String path) async {
    //Upload the file to firebase
    final StorageUploadTask uploadTask =
        _firestorageClient.ref().child(path).putFile(file);
    // Waits till the file is uploaded then stores the download url
    final StorageTaskSnapshot storageTaskSnapshot = await uploadTask.onComplete;

    return storageTaskSnapshot.ref.getDownloadURL(); //returns the download url
  }

  Future<DocumentReference> addToCollection(
    Map<String, dynamic> data,
    String path,
  ) async {
    final CollectionReference colRef = _firestoreClient.collection(path);
    return colRef.add(data);
  }

  // To verify is docId with username is present.
  // @path: should be col with document id which we wants.
  Future<bool> isDocExist(
    String path,
    String docId,
  ) async {
    final DocumentSnapshot docSnapShot =
        await _firestoreClient.collection(path).document(docId).get();
    return docSnapShot.exists;
  }

  // To set document data to document
  // @path : should be document level only
  Future<void> setDoc(
    String path,
    Map<String, dynamic> data,
  ) async {
    return _firestoreClient.document(path).setData(data);
  }

  /// To update document
  /// @params: [path]: must be collection level
  /// @params: [docId]: must be exist documentID
  /// @params: [data]: must be map data
  /// @return: updated document
  Future<dynamic> updateDoc(
    String path,
    String docId,
    Map<String, dynamic> data,
  ) async {
    await _firestoreClient.collection(path).document(docId).updateData(data);
    return (await loadDoc('$path/$docId')).data;
  }

  // To to remove specific doc from collection
  // @path : should be doc level only
  Future<void> removeDoc(String path) async {
    return _firestoreClient.document(path).delete();
  }

  // To load single document from specific col with docId
  // @path: should be document level only
  Future<DocumentSnapshot> loadDoc(String path) {
    final DocumentReference docRef = _firestoreClient.document(path);
    return docRef.get();
  }

  // To check exist same value on document
  // @path: should be collectioin level only
  Future<bool> checkDocByKey(
    String path,
    String key,
    dynamic value,
  ) async {
    final QuerySnapshot querySnap = await _firestoreClient
        .collection(path)
        .where(key, isEqualTo: value)
        .getDocuments();
    return querySnap.documents.isNotEmpty;
  }

  // load stream collection base on provided search criteria
  Stream<QuerySnapshot> loadCollectionStream({
    String path,
    String searchKey,
    String searchValue,
    String searchArrayKey,
    String searchArrayValue,
    bool noOrdered,
  }) {
    final CollectionReference colRef = _firestoreClient.collection(path);
    if (searchKey != null && searchValue != null) {
      if (noOrdered != null && !!noOrdered) {
        return colRef.where(searchKey, isEqualTo: searchValue).snapshots();
      } else {
        return colRef
            .orderBy('createdAt', descending: true)
            .where(searchKey, isEqualTo: searchValue)
            .snapshots();
      }
    } else if (searchArrayKey != null && searchArrayValue != null) {
      return colRef
          .orderBy('createdAt', descending: true)
          .where(searchArrayKey, arrayContains: searchArrayValue)
          .snapshots();
    } else if (noOrdered != null && !!noOrdered) {
      return colRef.snapshots();
    }
    return colRef.orderBy('createdAt', descending: true).snapshots();
  }

  /// load stream collection base on modern search criteria
  /// search params format:
  /// type: List of Map<String, dynamic>
  /// key: field name, opt: the operator for query, value: expected value for search
  /// If you want to ignore search params, just set value to []
  Stream<QuerySnapshot> loadCollectionStreamCustom([
    String path,
    List<Map<String, dynamic>> conditions,
    List<Map<String, dynamic>> orders,
  ]) {
    final CollectionReference colRef = _firestoreClient.collection(path);
    Query query = colRef;
    if (conditions.isNotEmpty) {
      for (var i = 0; i < conditions.length; i++) {
        query = queryWhere(query, conditions[i]);
      }
    }
    if (orders.isNotEmpty) {
      for (var i = 0; i < orders.length; i++) {
        query = query.orderBy(
          orders[i]['field'],
          descending: orders[i]['isDescending'] as bool,
        );
      }
    }
    return query.snapshots();
  }

  /// To get query with where
  /// @params: [query]: must be collection reference or query
  /// @params: [condition]: Map data, [key]: document field, [opt]: operator, [value]: filter value
  Query queryWhere(
    Query query,
    Map<String, dynamic> condition,
  ) {
    switch (condition['opt'] as String) {
      case '==':
        return query.where(condition['key'], isEqualTo: condition['value']);
        break;
      case '<':
        return query.where(condition['key'], isLessThan: condition['value']);
        break;
      case '<=':
        return query.where(condition['key'],
            isLessThanOrEqualTo: condition['value']);
        break;
      case '>':
        return query.where(condition['key'], isGreaterThan: condition['value']);
        break;
      case '>=':
        return query.where(condition['key'],
            isGreaterThanOrEqualTo: condition['value']);
        break;
      case 'in':
        return query.where(condition['key'], arrayContains: condition['value']);
        break;
      case 'isNull':
        return query.where(condition['key'], isNull: true);
        break;
      case 'isNotNull':
        return query.where(condition['key'], isNull: false);
        break;
      default:
        return query.where(condition['key'], isEqualTo: condition['value']);
    }
  }

  // To load multiple document from collection base on query
  // @path: should be col level only
  Future<QuerySnapshot> loadCollection(String path) {
    final CollectionReference colRef = _firestoreClient.collection(path);
    return colRef.getDocuments();
  }

  Future<QuerySnapshot> loadCollectionCustom([
    String path,
    List<Map<String, dynamic>> wheres,
    List<Map<String, dynamic>> orders,
  ]) {
    final CollectionReference colRef = _firestoreClient.collection(path);
    Query query = colRef;
    if (wheres != null) {
      for (var i = 0; i < wheres.length; i++) {
        query = queryWhere(query, wheres[i]);
      }
    }
    if (orders != null) {
      for (var i = 0; i < orders.length; i++) {
        query = query.orderBy(
          orders[i]['field'],
          descending: orders[i]['descending'] as bool,
        );
      }
    }
    return query.getDocuments();
  }

  Stream<Map<String, dynamic>> loadDocStream(String path) {
    final DocumentReference docRef = _firestoreClient.document(path);
    return docRef.snapshots().asyncMap((doc) => doc.data);
  }
}
